class DeletableMixin:
    '''Interface for classes that can be edited and deleted.'''

    def delete(self):
        '''Delete the object.

        Example usage:

        .. code:: python

            bot = credmgr.bot(name='name')
            bot.delete()

        '''
        self._credmgr.delete(f'{self._path}/{self.id}')
        del self


from . import RedditApp
from .utils import resolveModelFromInput, resolveUser
from ..mixins import BaseModel, DeletableMixin, EditableMixin, OwnerMixin


class UserVerification(BaseModel, DeletableMixin, EditableMixin, OwnerMixin):
    _attrTypes = {
        **BaseModel._attrTypes, 'id': 'int', 'user_id': 'str', 'redditor': 'str', 'reddit_app_id': 'int', 'extra_data': 'str', 'owner_id': 'int'
    }

    _editableAttrs = ['userId', 'redditor', 'redditAppId', 'extraData']
    _path = '/user_verifications'
    _credmgrCallable = 'userVerification'
    _nameAttr = 'userId'
    _canFetchByName = True
    _getByNamePath = 'get_redditor'
    _fetchNameAttr = 'user_id'

    def __init__(self, credmgr, id=None, userId=None, redditor=None, redditAppId=None, extraData=None, ownerId=None):
        super().__init__(credmgr, id)
        self.userId = userId
        self.redditAppId = redditAppId
        self.redditApp = self._credmgr.redditApp(self.redditAppId)
        if redditor:
            self.redditor = redditor
        if extraData:
            self.extraData = extraData
        self.ownerId = ownerId

    @staticmethod
    @resolveUser()
    def _create(_credmgr, userId, redditApp, redditor=None, extraData=None, owner=None):
        '''Create a new User Verification

        **PERMISSIONS: At least Active user is required.**

        User Verifications for verifying a redditor with a User ID

        :param str userId: User ID to associate Redditor with (required)
        :param Union[RedditApp,int,str] redditApp: Reddit app the User Verification is for (required)
        :param str redditor: Redditor the User Verification is for. This is not usually set manually.
        :param dict extraData: Extra JSON data to include with verification
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to create for other users.
        :return: UserVerification
        '''

        data = {'user_id': userId, 'reddit_app_id': resolveModelFromInput(_credmgr, RedditApp, redditApp)}
        if redditor:
            data['redditor'] = redditor
        if extraData:
            data['extra_data'] = extraData
        if owner:
            data['owner_id'] = owner
        return _credmgr.post('/user_verifications', data=data)


from credmgr.models.utils import camelToSnake


class EditableMixin:
    _editableAttrs = []

    def edit(self, **kwargs):
        payload = []

        for attr in self._editableAttrs:
            if attr in kwargs:
                payload.append({'op': 'replace', 'path': f'/{camelToSnake(attr)}', 'value': kwargs[attr]})
        self.__dict__.update(self._credmgr.patch(f'{self._path}/{self.id}', data=payload).__dict__)
        return self


import base64

import praw

from .utils import resolveModelFromInput, resolveUser
from ..mixins import BaseApp


class RedditApp(BaseApp):
    _attrTypes = {
        **BaseApp._attrTypes,
        'client_id': 'str',
        'client_secret': 'str',
        'short_name': 'str',
        'app_description': 'str',
        'user_agent': 'str',
        'app_type': 'str',
        'redirect_uri': 'str',
        'state': 'str'
        }

    _editableAttrs = BaseApp._editableAttrs + ['client_id', 'client_secret', 'short_name', 'app_description', 'user_agent', 'app_type', 'redirect_uri']
    _path = '/reddit_apps'
    _credmgrCallable = 'redditApp'

    def __init__(self, credmgr, id=None, appName=None, clientId=None, clientSecret=None, shortName=None, appDescription=None, userAgent=None,
                 appType=None, redirectUri=None, state=None, enabled=None, ownerId=None):
        super(RedditApp, self).__init__(credmgr, id, appName, enabled, ownerId)
        self.clientId = clientId
        self.clientSecret = clientSecret
        self.shortName = shortName
        self.appDescription = appDescription
        self.userAgent = userAgent
        self.appType = appType
        self.redirectUri = redirectUri
        self.state = state
        self.enabled = enabled
        self.ownerId = ownerId

    @staticmethod
    @resolveUser()
    def _create(_credmgr, appName, clientId, userAgent, appType, redirectUri, clientSecret, shortName, appDescription, enabled, owner=None):
        '''Create a new Reddit App

        **PERMISSIONS: At least Active user is required.**   Reddit Apps are used for interacting with reddit

        :param str appName: (required)
        :param str clientId: Client ID of the Reddit App (required)
        :param str userAgent: User agent used for requests to Reddit's API (required)
        :param str appType: Type of the app. One of `web`, `installed`, or `script` (required)
        :param str redirectUri: Redirect URI for Oauth2 flow. Defaults to user set redirect uri (required)
        :param str clientSecret: Client secret of the Reddit App
        :param str shortName: Short name of the Reddit App
        :param str appDescription: Description of the Reddit App
        :param bool enabled: Allows the app to be used
        :param int owner: Owner of the app. Requires Admin to create for other users.
        :return: RedditApp
        '''
        data = {'app_name': appName, 'client_id': clientId, 'user_agent': userAgent, 'app_type': appType, 'redirect_uri': redirectUri}
        if clientSecret:
            data['client_secret'] = clientSecret
        if shortName:
            data['short_name'] = shortName
        if appDescription:
            data['app_description'] = appDescription
        if enabled:
            data['enabled'] = enabled
        if owner:
            data['ownerId'] = owner
        return _credmgr.post('/reddit_apps', data=data)

    def reddit(self, redditor=None) -> praw.reddit:
        refreshToken = None
        if redditor:
            refreshToken = self._credmgr.refreshToken(redditor, self.id)
            if refreshToken:
                refreshToken = refreshToken.refreshToken
        return praw.Reddit(client_id=self.clientId, client_secret=self.clientSecret, user_agent=self.userAgent, redirect_uri=self.redirectUri, refresh_token=refreshToken)

    def genAuthUrl(self, scopes=None, permanent=False, userVerification=None):
        '''Generates an URL for users to verify or authenciate their Reddit account

        :param Union[list,str] scopes: List of scopes needed. Pass ``'all'`` for all scopes. The ``identity`` scope will always be included. (default: ``['identity']``)
        :param bool permanent: Determines if a refresh token will be provided. (default: ``False``)
        :param Union[UserVerification,int,str] userVerification: Link to a :class:`.UserVerification` object. Accepted values are:

            - An :class:``.UserVerification`` object
            - An :class:``.UserVerification`` id
            - An ``userId`` of a :class:``.UserVerification`` record.
         If a :class:`.UserVerification` record does not exist, one will be created.
        :return str: Auth URL
        '''
        from credmgr.models import UserVerification
        if scopes is None:
            scopes = ['identity']
        elif scopes == 'all':
            scopes = ['*']
        if not 'identity' in scopes and scopes != ['*']:
            scopes.append('identity')
        if permanent:
            duration = 'permanent'
        else:
            duration = 'temporary'
        userVerification = resolveModelFromInput(self._credmgr, UserVerification, userVerification, 'userId')
        if not userVerification and userVerification:
            userVerification = self._credmgr.userVerification.create(userVerification, self.id)
        if userVerification:
            state = base64.urlsafe_b64encode(f'{self.state}:{userVerification}'.encode()).decode()
        else:
            state = self.state
        return self.reddit().auth.url(scopes, state, duration)


from urllib.parse import urljoin

from requests import Session

from . import models
from .auth import ApiTokenAuth
from .exceptions import InitializationError
from .requestor import Requestor
from .serializer import Serializer


User = models.User
Bot = models.Bot
RedditApp = models.RedditApp
RefreshToken = models.RefreshToken
UserVerification = models.UserVerification
SentryToken = models.SentryToken
DatabaseCredential = models.DatabaseCredential

def setOwner(ownerId, data):
    if ownerId:
        data['ownerId'] = ownerId

class CredentialManager(object):
    '''The CredentialManager class provides convenient access to CredentialManager's API.

    Instances of this class are the gateway to interacting with CredentialManager's API
    through CredentialManager. The canonical way to obtain an instance of this class is via:


    .. code-block:: python

        import CredentialManager
        credmgr = CredentialManager.client(apiToken='LIqbGjAeep3Ws5DH3LOEQPmw8UZ6ek')


    '''
    _default = None
    _endpoint = '/api/v1'

    def __init__(self, host='https://credmgr.jesassn.org', apiToken=None, username=None, password=None, sessionClass=None, sessionKwargs={}):
        '''Initialize a CredentialManager instance.

        :param str host: Hostname to use for interacting with the API (default: ``https://credmgr.jesassn.org``)
        :param str apiToken: API Token used for authentication (Note: cannot be used when ``username`` and ``password`` params are passed)
        :param str username: Username to use for interacting with the API (Note: Cannot be used when ``apiToken`` param is passed)
        :param str password: Password to use for interacting with the API (Note: Cannot be used when ``apiToken`` param is passed)
        :param Session sessionClass: A Session class that will be used to create a requestor. If not set, use ``requests.Session`` (default: None).
        :param dict sessionKwargs: Dictionary with additional keyword arguments used to initialize the session (default: None).

        .. warning::
             Using an API Token instead of a username/password is strongly recommended!

        The ``sessionClass`` and ``sessionKwargs`` allow for
        customization of the session :class:`.CredentialManager` will use. This allows,
        e.g., easily adding behavior to the requestor or wrapping its
        |Session|_ in a caching layer. Example usage:

        .. |Session| replace:: ``Session``
        .. _Session: https://2.python-requests.org/en/master/api/#requests.Session

        .. code-block:: python

           import json, betamax, requests

           class JSONDebugRequestor(Requestor):
               def request(self, *args, **kwargs):
                   response = super().request(*args, **kwargs)
                   print(json.dumps(response.json(), indent=4))
                   return response

           mySession = betamax.Betamax(requests.Session())
           reddit = CredentialManager.client(..., sessionClass=JSONDebugRequestor, sessionKwargs={'session': mySession})
        '''
        self._host = urljoin(host, self._endpoint)
        if all([apiToken, username, password]):
            raise InitializationError('Either apiToken or username/password pair can be passed, both cannot be passed')
        if apiToken:
            self._auth = ApiTokenAuth(apiToken)
        elif username and password:
            self._auth = (username, password)
        else:
            raise InitializationError('apiToken or an username/password pair must be passed')

        self._requestor = Requestor(self._host, self._auth, sessionClass, **sessionKwargs)
        self.serializer = Serializer(self)
        self.currentUser: User = self.currentUser()
        self.userDefaults = self.currentUser.defaultSettings
        self.getUserDefault = lambda key, default: self.userDefaults.get(key, default)
        self.user = models.UserHelper(self)
        '''An instance of :class:`.UserHelper`.

        Provides the interface for interacting with :class:`.User`.
        For example to get a ``user`` with ``id`` of ``1`` do:
        
        .. code-block:: python
            user = credmgr.user(1)
            print(user.id)
        
        To create a ``user`` do:
        
        ..code-block:: python
            user = credmgr.user.create(**userKwargs)
        
        See :meth:`~.UserHelper.create` for the required params.
        '''
        self.bot = models.BotHelper(self)
        '''An instance of :class:`.BotHelper`.

        Provides the interface for interacting with :class:`.Bot`.
        For example to get a ``bot`` with ``id`` of ``1`` do:
        
        .. code-block:: python
            bot = credmgr.bot(1)
            print(bot.id)
        
        To create a ``bot`` do:
        
        ..code-block:: python
            bot = credmgr.bot.create(**botKwargs)
        
        See :meth:`~.BotHelper.create` for the required params.
        '''

        self.redditApp = models.RedditAppHelper(self)
        '''An instance of :class:`.RedditAppHelper`.

        Provides the interface for interacting with :class:`.RedditApp`.
        For example to get a ``redditApp`` with ``id`` of ``1`` do:
        
        .. code-block:: python
            redditApp = credmgr.redditApp(1)
            print(redditApp.id)
        
        To create a ``redditApp`` do:
        
        ..code-block:: python
            redditApp = credmgr.redditApp.create(**redditAppKwargs)
        
        See :meth:`~.RedditAppHelper.create` for the required params.
        '''

        self.refreshToken = models.RefreshTokenHelper(self)
        '''An instance of :class:`.RefreshTokenHelper`.

        Provides the interface for interacting with :class:`.RefreshToken`.
        For example to get a ``refreshToken`` with ``id`` of ``1`` do:
        
        .. code-block:: python
            refreshToken = credmgr.refreshToken(1)
            print(refreshToken.id)
        
        .. note::
            Refresh tokens cannot be manually created.
        '''

        self.userVerification = models.UserVerificationHelper(self)
        '''An instance of :class:`.UserVerificationHelper`.

        Provides the interface for interacting with :class:`.UserVerification`.
        For example to get a ``userVerification`` with ``id`` of ``1`` do:
        
        .. code-block:: python
            userVerification = credmgr.userVerification(1)
            print(userVerification.id)
        
        To create a ``userVerification`` do:
        
        ..code-block:: python
            userVerification = credmgr.userVerification.create(**userVerificationKwargs)
        
        See :meth:`~.UserVerificationHelper.create` for the required params.
        '''

        self.sentryToken = models.SentryTokenHelper(self)
        '''An instance of :class:`.SentryTokenHelper`.

        Provides the interface for interacting with :class:`.SentryToken`.
        For example to get a ``sentryToken`` with ``id`` of ``1`` do:
        
        .. code-block:: python
            sentryToken = credmgr.sentryToken(1)
            print(sentryToken.id)
        
        To create a ``sentryToken`` do:
        
        ..code-block:: python
            sentryToken = credmgr.sentryToken.create(**sentryTokenKwargs)
        
        See :meth:`~.SentryTokenHelper.create` for the required params.
        '''

        self.databaseCredential = models.DatabaseCredentialHelper(self)
        '''An instance of :class:`.DatabaseCredentialHelper`.

        Provides the interface for interacting with :class:`.DatabaseCredential`.
        For example to get a ``databaseCredential`` with ``id`` of ``1`` do:
        
        .. code-block:: python
            databaseCredential = credmgr.databaseCredential(1)
            print(databaseCredential.id)
        
        To create a ``databaseCredential`` do:
        
        ..code-block:: python
            databaseCredential = credmgr.databaseCredential.create(**databaseCredentialKwargs)
        
        See :meth:`~.DatabaseCredentialHelper.create` for the required params.
        '''

    def users(self, batchSize=10, limit=None):
        return User(self).listItems(batchSize=batchSize, limit=limit)

    def bots(self, batchSize=20, limit=None, owner=None):
        return Bot(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def redditApps(self, batchSize=20, limit=None, owner=None):
        return RedditApp(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def refreshTokens(self, batchSize=20, limit=None, owner=None):
        return RefreshToken(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def userVerifications(self, batchSize=20, limit=None, owner=None):
        return UserVerification(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def sentryTokens(self, batchSize=20, limit=None, owner=None):
        return SentryToken(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def databaseCredentials(self, batchSize=20, limit=None, owner=None):
        return DatabaseCredential(self).listItems(batchSize=batchSize, limit=limit, owner=owner)

    def currentUser(self) -> User:
        return self.get('/users/me')

    def get(self, path, params=None):
        return self.serializer.deserialize(self._requestor.request(path, 'GET', params=params))

    def post(self, path, data):
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        return self.serializer.deserialize(self._requestor.request(path, 'POST', data=data, headers=headers))

    def patch(self, path, data):
        return self.serializer.deserialize(self._requestor.request(path, 'PATCH', json=data))

    def delete(self, path):
        return self._requestor.request(path, 'DELETE')


from credmgr.mixins import BaseModel, DeletableMixin, OwnerMixin


class RefreshToken(BaseModel, DeletableMixin, OwnerMixin):
    _attrTypes = {
        **BaseModel._attrTypes,
        'reddit_app_id': 'int',
        'redditor': 'str',
        'refresh_token': 'str',
        'owner_id': 'int',
        'scopes': 'str',
        'issued_at': 'datetime',
        'revoked': 'bool',
        'revoked_at': 'datetime'
    }

    _path = '/refresh_tokens'
    _credmgrCallable = 'refreshToken'
    _fetchNameAttr = {'redditor': 'redditor', 'redditAppId': 'reddit_app_id'}
    _getByNamePath = 'by_redditor'
    _canFetchByName = True
    _nameAttr = 'redditor'

    def __init__(self, credmgr, id=None, redditAppId=None, redditor=None, refreshToken=None, ownerId=None, scopes=None, issuedAt=None, revoked=False, revokedAt=None):
        super().__init__(credmgr, id)
        self.redditAppId = redditAppId
        self.redditApp = self._credmgr.redditApp(self.redditAppId)
        self.redditor = redditor
        self.refreshToken = refreshToken
        self.ownerId = ownerId
        self.scopes = scopes
        self.issuedAt = issuedAt
        self.revoked = revoked
        self.revokedAt = revokedAt


from requests.auth import AuthBase


class ApiTokenAuth(AuthBase):

    def __init__(self, apiToken):
        self.apiToken = apiToken

    def __call__(self, request):
        request.headers['X-API-TOKEN'] = self.apiToken
        return request


from credmgr.models.utils import camelToSnake


class ToggableMixin:
    _enabledAttr = 'enabled'

    def toggle(self, enabled):
        return self._credmgr.patch(f'{self._path}/{id.id}', body=[{'op': 'replace', 'path': f'/{camelToSnake(self._enabledAttr)}', 'value': enabled}])


import datetime
import re

import credmgr.models
from .exceptions import SerializerException


class Serializer(object):
    '''The serializer builds :class:`.BaseModel` objects.'''

    primitiveTypes = (float, bool, bytes, str, int)
    nativeTypeMapping = {
        'int': int,
        'str': str,
        'bool': bool,
        'datetime': datetime.datetime,
        'object': object,
        'dict': dict,
        'list': list
        }

    def __init__(self, credmgr):
        '''Initialize an Objector instance.

        :param credmgr: An instance of :class:`~.CredentialManager`.

        '''
        self._credmgr = credmgr

    @staticmethod
    def deserializeDatatime(string):
        '''Deserializes string to datetime.

        The string should be in iso8601 datetime format.

        :param string: str.
        :return: datetime.
        '''
        try:
            from dateutil.parser import parse
            return parse(string)
        except ImportError:
            return string
        except ValueError:
            raise SerializerException(f"Failed to parse `{string}` as datetime object")

    @staticmethod
    def deserializeDate(string):
        '''Deserializes string to date.

        :param string: str.
        :return: date.
        '''
        try:
            from dateutil.parser import parse
            return parse(string).date()
        except ImportError:
            return string
        except ValueError:
            raise SerializerException(f"Failed to parse `{string}` as date object")

    @staticmethod
    def deserializeObject(value):
        '''Return a original value.

        :return: object.
        '''
        return value

    @staticmethod
    def deserializePrimitive(data, objectType):
        '''Deserializes string to primitive type.

        :param data: str.
        :param objectType: class literal.

        :return: int, long, float, str, bool.
        '''
        try:
            return objectType(data)
        except UnicodeEncodeError:
            return str(data)
        except TypeError:
            return data

    def deserialize(self, response):
        '''Deserializes response into an object.

        :param response: Response object to be deserialized.

        :return: deserialized object.
        '''
        data = response.json()
        if isinstance(data, list) and all([i['resource_type'] == data[0]['resource_type'] for i in data]):
            if data:
                resourceType = f'list[{data[0]["resource_type"]}]'
            else:
                return data
        else:
            resourceType = data.get('resource_type', 'dict')
        return self.__deserialize(data, resourceType)

    def __deserialize(self, data, objectType):
        '''Deserializes dict, list, str into an object.

        :param data: dict, list or str.
        :param objectType: class literal, or string of class name.

        :return: object.
        '''
        if data is None:
            return None

        if type(objectType) == str:
            if objectType.startswith('list['):
                subKls = re.match(r'list\[(.*)\]', objectType).group(1)
                return [self.__deserialize(subData, subKls) for subData in data]

            if objectType.startswith('dict'):
                return {k: self.__deserialize(v, type(v)) for k, v in data.items()}

            # convert str to class
            if objectType in self.nativeTypeMapping:
                objectType = self.nativeTypeMapping[objectType]
            else:
                objectType = getattr(credmgr.models, objectType)

        if objectType in self.primitiveTypes:
            return self.deserializePrimitive(data, objectType)
        elif objectType == object:
            return self.deserializeObject(data)
        elif objectType == datetime.date:
            return self.deserializeDate(data)
        elif objectType == datetime.datetime:
            return self.deserializeDatatime(data)
        elif objectType == dict:
            return self.__deserialize(data, 'dict')
        else:
            return self.__deserializeModel(data, objectType)

    def __hasattr(self, object, name):
        return name in object.__class__.__dict__

    def __deserializeModel(self, data, objectType):
        '''Deserializes list or dict to model.

        :param data: dict, list.
        :param objectType: class literal.
        :return: model object.
        '''

        if not objectType._attrTypes:
            return data

        kwargs = {}
        snakeToCamel = lambda snakeCase: ''.join(snakeCase.split('_')[:1]+[i.capitalize() for i in snakeCase.split('_')[1:]])
        if objectType._attrTypes is not None:
            for attr, attrType in objectType._attrTypes.items():
                if data and attr in data and isinstance(data, (list, dict)):
                    value = data[attr]
                    kwargs[snakeToCamel(attr)] = self.__deserialize(value, attrType)
        try:
            instance = objectType(self._credmgr, **kwargs)
        except TypeError:
            for key, value in kwargs.items():
                setattr(objectType, key, value)
            instance = objectType
        if isinstance(instance, dict) and objectType._attrTypes is not None and isinstance(data, dict):
            for key, value in data.items():
                if key not in objectType._attrTypes:
                    instance[key] = value
        return instance


from credmgr.models.utils import resolveUser


class BaseModel(object):
    _attrTypes = {'id': 'int'}
    _path = None
    _nameAttr = None
    _canFetchByName = False
    _credmgrCallable = None
    _getByNamePath = 'by_name'
    _fetchNameAttr = 'app_name'

    def __init__(self, credmgr, id=None, name=None):
        self._credmgr = credmgr
        self.id = id
        self.fetched = False
        if self._nameAttr and self._canFetchByName:
            setattr(self, self._nameAttr, name)

    def __getattr__(self, attribute):
        '''Return the value of `attribute`.'''
        if not attribute.startswith('_'):
            if not self.fetched:
                self.fetch()
            return self.__dict__.get(attribute, None)
        else:
            return None

    def get(self, id):
        self.__dict__ = self._credmgr.get(f'{self._path}/{id}').__dict__
        self.fetched = True

    def getByName(self, name):
        data = {}
        if isinstance(self._fetchNameAttr, dict):
            for modelAttr, dataAttr in self._fetchNameAttr.items():
                data[dataAttr] = getattr(self, modelAttr)
        else:
            data[self._fetchNameAttr] = name
        self.__dict__ = self._credmgr.post(f'{self._path}/{self._getByNamePath}', data=data).__dict__
        self.fetched = True

    def fetch(self, byName=False):
        if byName and self._canFetchByName:
            self.getByName(getattr(self, self._nameAttr))
        else:
            self.get(self.id)

    @resolveUser()
    def listItems(self, batchSize=20, limit=None, owner=None):
        from credmgr.models.helpers import Paginator
        return Paginator(self._credmgr, self.__class__, batchSize=batchSize, limit=limit, owner=owner)

    def toDict(self):
        result = {}

        for attr, _ in self._attrTypes.items():
            attr = ''.join(attr.split('_')[:1] + [i.capitalize() for i in attr.split('_')[1:]])
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(lambda x: x.toDict() if hasattr(x, 'toDict') else x, value))
            elif hasattr(value, 'toDict'):
                result[attr] = value.toDict()
            elif isinstance(value, dict):
                result[attr] = dict(map(lambda item: (item[0], item[1].toDict()) if hasattr(item[1], 'toDict') else item, value.items()))
            else:
                result[attr] = value
        if issubclass(type(self), dict):
            for key, value in self.items():
                result[key] = value
        return result

    def __repr__(self):
        ownerRepr = ''
        if hasattr(self, 'owner'):
            ownerRepr = f', owner={self.owner.username!r}'
        return f'<{self.__class__.__name__} id={self.id}, {self._nameAttr}={getattr(self, self._nameAttr)!r}{ownerRepr}>'

    def __eq__(self, other):
        '''Returns true if both objects are equal'''
        if not isinstance(other, type(self)):
            return False

        return self.__dict__ == other.__dict__


from .bot import Bot
from .database_credential import DatabaseCredential
from .reddit_app import RedditApp
from .refresh_token import RefreshToken
from .sentry_token import SentryToken
from .user import User
from .user_verification import UserVerification


_ = None
del _
from .helpers import BotHelper, RedditAppHelper, UserHelper, UserVerificationHelper, SentryTokenHelper, DatabaseCredentialHelper, RefreshTokenHelper


from .utils import resolveUser
from ..mixins import BaseApp


class DatabaseCredential(BaseApp):
    _attrTypes = {
        **BaseApp._attrTypes,
        'database_username': 'str',
        'database_host': 'str',
        'database': 'str',
        'database_flavor': 'str',
        'database_port': 'int',
        'database_password': 'str',
        'use_ssh': 'bool',
        'ssh_host': 'str',
        'ssh_port': 'int',
        'ssh_username': 'str',
        'ssh_password': 'str',
        'use_ssh_key': 'bool',
        'private_key': 'str',
        'private_key_passphrase': 'str'
    }

    _editableAttrs = BaseApp._editableAttrs + ['databaseUsername', 'databaseHost', 'database', 'databaseFlavor', 'databasePort', 'databasePassword', 'useSsh', 'sshHost', 'sshPort',
                                               'sshUsername', 'sshPassword', 'useSshKey', 'privateKey', 'privateKeyPassphrase']
    _path = '/database_credentials'
    _credmgrCallable = 'databaseCredential'

    def __init__(self, credmgr, id=None, appName=None, databaseUsername=None, databaseHost=None, database=None, databaseFlavor=None, databasePort=None, databasePassword=None,
            useSsh=None, sshHost=None, sshPort=None, sshUsername=None, sshPassword=None, useSshKey=None, privateKey=None, privateKeyPassphrase=None, enabled=None, ownerId=None):
        super().__init__(credmgr, id, appName, enabled, ownerId)
        self.databaseUsername = databaseUsername
        self.databaseHost = databaseHost
        self.database = database
        self.databasePort = databasePort
        self.enabled = enabled
        self.ownerId = ownerId
        if databaseFlavor:
            self.databaseFlavor = databaseFlavor
        if databasePassword:
            self.databasePassword = databasePassword
        if useSsh:
            self.useSsh = useSsh
        if sshHost:
            self.sshHost = sshHost
        if sshPort:
            self.sshPort = sshPort
        if sshUsername:
            self.sshUsername = sshUsername
        if sshPassword:
            self.sshPassword = sshPassword
        if useSshKey:
            self.useSshKey = useSshKey
        if privateKey:
            self.privateKey = privateKey
        if privateKeyPassphrase:
            self.privateKeyPassphrase = privateKeyPassphrase

    @staticmethod
    @resolveUser()
    def _create(_credmgr, appName, databaseFlavor='postgres', database='postgres', databaseHost='localhost', databasePort=5432, databaseUsername='postgres', databasePassword=None,
            useSsh=False, sshHost=None, sshPort=None, sshUsername=None, sshPassword=None, useSshKey=False, privateKey=None, privateKeyPassphrase=None, enabled=True, owner=None):
        '''Create a new Database Credential

        **PERMISSIONS: At least Active user is required.**

        Database Credentials are used for..ya know..databases

        :param str appName: Name of the Database Credential (required)
        :param str databaseFlavor: Type of database, (default: ``postgres``)
        :param str database: Working database to use, (default: ``postgres``)
        :param str databaseHost: Database server address, (default: ``localhost``)
        :param int databasePort: Port the database server listens on, (default: ``5432``)
        :param str databaseUsername: Username to use to connect to the database
        :param str databasePassword: Password to use to connect to the database
        :param bool useSsh: Determines if the database will be connected to through a tunnel
        :param str sshHost: The address of the server that the SSH tunnel will connect to
        :param str sshPort: The port the SSH tunnel will use
        :param str sshUsername: Username for the SSH tunnel
        :param str sshPassword: Password for the SSH tunnel
        :param bool useSshKey: Allows the credentials to be used
        :param str privateKey: SSH private key. Note: No validation will be performed.
        :param str privateKeyPassphrase: Passphrase for the SSH key
        :param bool enabled: Allows the credentials to be used
        :param Union[User,int,str] owner: Owner of the app. Requires Admin to create for other users.
        :return: DatabaseCredential
        '''

        data = {}
        if databaseFlavor:
            data['database_flavor'] = databaseFlavor
        if database:
            data['database'] = database
        if databaseHost:
            data['database_host'] = databaseHost
        if databasePort:
            data['database_port'] = databasePort
        if databaseUsername:
            data['database_username'] = databaseUsername
        if databasePassword:
            data['database_password'] = databasePassword
        if useSsh:
            data['use_ssh'] = useSsh
        if sshHost:
            data['ssh_host'] = sshHost
        if sshPort:
            data['ssh_port'] = sshPort
        if sshUsername:
            data['ssh_username'] = sshUsername
        if sshPassword:
            data['ssh_password'] = sshPassword
        if useSshKey:
            data['use_ssh_key'] = useSshKey
        if privateKey:
            data['private_key'] = privateKey
        if privateKeyPassphrase:
            data['private_key_passphrase'] = privateKeyPassphrase
        if enabled:
            data['enabled'] = enabled
        if owner:
            data['owner_id'] = owner
        return _credmgr.post('/database_credentials', data={'app_name': appName, **data})


class CredentialManagerException(Exception):
    '''The base CredentialManager Exception that all other exception classes extend.'''

class InitializationError(CredentialManagerException):
    '''Invalid initialization paramaters were used.'''

class SerializerException(CredentialManagerException):
    '''Serializaton/deserialzation error occured.'''

class APIException(CredentialManagerException):
    '''Base exception class'''

    def __init__(self, response):
        self.url = response.url
        self.status = response.status_code
        self.reason = response.reason
        self.body = response.text
        self.headers = response.headers

    def __str__(self):
        '''Custom error messages for exception'''
        errorMessage = f'\n{self.status}: {self.url}\n{self.reason}'
        if self.headers:
            errorMessage += f'\nResponse headers: {self.headers}'
        if self.body:
            errorMessage += f'\nResponse body: {self.body}'

        return errorMessage

class Conflict(APIException):
    '''Made a conflicting request.'''

class Forbidden(APIException):
    '''Made request to a resource without required permissions.'''

class InvalidJSON(APIException):
    '''Response returned invalid JSON.'''

class InvalidRequest(APIException):
    '''Made request with invalid parameters.'''

class NotFound(APIException):
    '''Could not find requested resource.'''

class ServerError(APIException):
    '''Error occured on server during request.'''

class Unauthorized(APIException):
    '''Made request to resource without authenticating.'''


import logging

import requests
from requests import codes

from .const import __version__
from .exceptions import Conflict, Forbidden, InvalidJSON, InvalidRequest, NotFound, ServerError, Unauthorized


log = logging.getLogger(__package__)

def urljoin(base, path):
    if base.endswith('/'):
        base = base[:-1]
    if not path.startswith('/'):
        path = f'/{path}'
    return f'{base}{path}'

class Requestor(object):
    '''Requestor provides an interface to HTTP requests.'''
    _exceptionMapping = {
        400: InvalidRequest,
        401: Unauthorized,
        403: Forbidden,
        404: NotFound,
        409: Conflict,
        422: ServerError,
        500: ServerError,
        502: ServerError,
        503: ServerError,
        504: ServerError,
        520: ServerError,
        522: ServerError
        }
    _successCodes = [200, 201, 202, 204]

    def __init__(self, credmgrUrl, auth, session=None, **sessionKwargs):
        '''Create an instance of the Requestor class.

        :param str credmgrUrl: Url used to make API requests
        :param auth: An auth tuple or a class that subclasses requests.auth
        :param Session session: (Optional) A session to handle requests, compatible with requests.Session(). (Default: None)
        '''
        self._baseUrl = credmgrUrl
        self._session = session or requests.Session()
        self._session.auth = auth
        for key, value in sessionKwargs.items():
            setattr(self._session, key, value)
        self._session.headers['User-Agent'] = f'credmgr/{__version__}'

    def __getattr__(self, attribute):
        '''Pass all undefined attributes to the _session attribute.'''
        if attribute.startswith('__'):
            raise AttributeError
        return getattr(self._session, attribute)

    @staticmethod
    def _logRequest(data, method, params, url):
        log.info(f'Data: {data}')
        log.info(f'Request: {method} {url}')
        log.info(f'Query Parameters: {params}')

    def close(self):
        '''Call close on the underlying session.'''
        return self._session.close()

    def request(self, path, method, data=None, params=None, **kwargs):
        url = urljoin(self._baseUrl, path)
        self._logRequest(data, method, params, url)
        response = self._session.request(method, url, params, data=data, **kwargs)
        if response.status_code in self._exceptionMapping:
            raise self._exceptionMapping[response.status_code](response)
        elif response.status_code == codes['no_content']:
            return
        assert (response.status_code in self._successCodes), f'Unexpected status code: {response.status_code}'
        if response.headers.get('content-length') == '0':
            return ''
        try:
            response.json()
        except ValueError:
            raise InvalidJSON(response)
        log.info(f'Response: {response.status_code} ({response.headers.get("content-length")} bytes)')
        return response


from ..mixins import BaseModel, DeletableMixin, EditableMixin, ToggableMixin


class User(BaseModel, DeletableMixin, EditableMixin, ToggableMixin):
    _attrTypes = {
        **BaseModel._attrTypes,
        'id': 'int',
        'username': 'str',
        'is_active': 'bool',
        'is_regular_user': 'bool',
        'is_admin': 'bool',
        'default_settings': 'dict(str, str)',
        'reddit_username': 'str',
        'created': 'datetime',
        'updated': 'datetime'
    }

    _path = '/users'
    _credmgrCallable = 'user'
    _nameAttr = 'username'
    _enabledAttr = 'isActive'
    _canFetchByName = True
    _fetchNameAttr = _nameAttr

    def __init__(self, credmgr, id=None, username=None, isActive=None, isRegularUser=None, isAdmin=None, defaultSettings=None, redditUsername=None, created=None, updated=None):
        super().__init__(credmgr, id)
        if username:
            self.username = username
        if isActive is not None:
            self.isActive = isActive
        if isRegularUser is not None:
            self.isRegularUser = isRegularUser
        if isAdmin is not None:
            self.isAdmin = isAdmin
        if defaultSettings is not None:
            self.defaultSettings = defaultSettings
        if redditUsername is not None:
            self.redditUsername = redditUsername
        if created is not None:
            self.created = created
        if updated is not None:
            self.updated = updated

    @staticmethod
    def _create(_credmgr, username, password, defaultSettings=None, redditUsername=None, isAdmin=False, isActive=True, isRegularUser=True, isInternal=False):
        '''Create a new User

        **PERMISSIONS: Admin role is required.**

        :param str username: Username for new user (Example: ```spaz```) (required)
        :param str password: Password for new user (Example: ```supersecurepassword```) (required)
        :param str defaultSettings: Default values to use for new apps (Example: ```{"databaseFlavor": "postgres", "databaseHost": "localhost"}```)
        :param str redditUsername: User's Reddit username (Example: ```LilSpazJoekp```)
        :param bool isAdmin: Is the user an admin? Allows the user to see all objects and create users (Default: ``false``)
        :param bool isActive: Is the user active? Allows the user to sign in (Default: ``true``)
        :param bool isRegularUser: (Internal use only)
        :param bool isInternal: (Internal use only)
        :return: User

        '''
        additionalParams = {}
        if defaultSettings:
            additionalParams['default_settings'] = defaultSettings
        if isAdmin:
            additionalParams['is_admin'] = isAdmin
        if isActive:
            additionalParams['is_active'] = isActive
        if isRegularUser:
            additionalParams['is_regular_user'] = isRegularUser
        if isInternal:
            additionalParams['is_internal'] = isInternal
        if redditUsername:
            additionalParams['reddit_username'] = redditUsername
        return _credmgr.post('/users', data={'username': username, 'password': password, **additionalParams})


import re

from credmgr.exceptions import NotFound


reCamelToSnake = re.compile(r"([a-z0-9](?=[A-Z])|[A-Z](?=[A-Z][a-z]))")

def camelToSnake(name: str) -> str:
    """Convert `name` from camelCase to snake_case."""
    return reCamelToSnake.sub(r"\1_", name).lower()

def resolveUser(userAttr='owner', returnAttr='id'):
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            from . import User
            value = None
            user = kwargs.get(userAttr, None)
            if user:
                if isinstance(user, User):
                    value = getattr(user, returnAttr)
                elif isinstance(user, int):
                    value = user
                elif isinstance(user, str):
                    foundUser = self._credmgr.user(user)
                    if foundUser:
                        value = getattr(foundUser, returnAttr)
                kwargs[userAttr] = value
            return func(self, *args, **kwargs)

        return wrapper

    return decorator

def resolveModelFromInput(credmgr, model, input, returnAttr='id'):
    value = None
    if isinstance(input, model):
        value = getattr(input, returnAttr)
    elif isinstance(input, int):
        value = input
    elif isinstance(input, str):
        try:
            foundItem = getattr(credmgr, model._credmgrCallable)(input)
        except NotFound:
            foundItem = None
        if foundItem:
            value = getattr(foundItem, returnAttr)
    return value


from ..mixins import BaseModel, DeletableMixin, EditableMixin, OwnerMixin, ToggableMixin


class BaseApp(BaseModel, DeletableMixin, EditableMixin, ToggableMixin, OwnerMixin):
    _attrTypes = {**BaseModel._attrTypes, 'app_name': 'str', 'enabled': 'bool', 'owner_id': 'int'}
    _nameAttr = 'appName'
    _editableAttrs = ['appName', 'enabled']


    def __init__(self, credmgr, id=None, appName=None, enabled=None, ownerId=None):
        super(BaseApp, self).__init__(credmgr, id)
        if appName:
            self.appName = appName
        if enabled:
            self.enabled = enabled
        if ownerId:
            self.ownerId = ownerId


import os
from os import path
import re

from setuptools import find_packages, setup


projectName = 'credmgr'

root = path.abspath(path.dirname(__file__))

with open(path.join(root, projectName, "const.py")) as f:
    __version__ = re.search(r"__version__ = '(.+)'", f.read()).group(1)

with open(path.join(root, "README.rst")) as f:
    longDescription = f.read()

requires = ['setuptools', 'python-dateutil', 'requests', 'praw', 'prawcore']
# testsRequires = [
#     'coverage',
#     'pytest',
#     'betamax',
#     'sphinx'
# ]

setup(
    name=projectName,
    author='Lil_SpazJoekp',
    author_email='spaz@jesassn.org',
    license='Proprietary',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Operating System :: OS Independent',
        'License :: Other/Proprietary License'
    ],
    python_requires='>=3.6',
    description='Credential Manager API Client',
    include_package_data=True,
    install_requires=requires,
    long_description=longDescription,
    packages=find_packages(exclude=['tests', 'tests.*']),
    test_suite='tests', # tests_require=testsRequires,
    url='https://credmgr.jesassn.org',
    version=__version__
)


from .utils import resolveUser
from ..mixins import BaseApp


class SentryToken(BaseApp):
    _attrTypes = {**BaseApp._attrTypes, 'dsn': 'str'}
    _editableAttrs = BaseApp._editableAttrs + ['dsn']
    _path = '/sentry_tokens'
    _credmgrCallable = 'sentryToken'

    def __init__(self, credmgr, id=None, appName=None, dsn=None, enabled=None, ownerId=None):
        super().__init__(credmgr, id, appName, enabled, ownerId)
        self.dsn = dsn

    @staticmethod
    @resolveUser()
    def _create(_credmgr, appName=None, dsn=None, owner=None):
        '''Create a new Sentry Token

        **PERMISSIONS: At least Active user is required.**

        Sentry Tokens are used for logging and error reporting in applications

        :param str appName: Name of the Sentry Token (required)
        :param str dsn: DSN of the Sentry Token (required)
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to create for other users.
        :return: SentryToken
        '''
        return _credmgr.post('/sentry_tokens', data={'app_name': appName, 'dsn': dsn})


from .utils import camelToSnake, resolveUser
from ..mixins import BaseApp


class Bot(BaseApp):
    _attrTypes = {
        **BaseApp._attrTypes, 'reddit_app': 'RedditApp', 'sentry_token': 'SentryToken', 'database_credential': 'DatabaseCredential'
    }
    _editableAttrs = BaseApp._editableAttrs + ['reddit_app_id', 'sentry_token_id', 'database_credential_id']
    _path = '/bots'
    _credmgrCallable = 'bot'
    _canFetchByName = True

    def __init__(self, credmgr, id=None, appName=None, enabled=None, ownerId=None, redditApp=None, sentryToken=None, databaseCredential=None):
        super(Bot, self).__init__(credmgr, id, appName, enabled, ownerId)
        self.redditApp = redditApp
        self.sentryToken = sentryToken
        self.databaseCredential = databaseCredential

    @staticmethod
    @resolveUser()
    def _create(_credmgr, name, redditApp, sentryToken, databaseCredential, owner=None):
        '''Create a new Bot

        **PERMISSIONS: At least Active user is required.**   Bots are used for grouping apps into a single request

        :param str name: Name of the Bot (required)
        :param Union[RedditApp,int] redditApp: Reddit App the bot will use
        :param Union[SentryToken,int] sentryToken: Sentry Token the bot will use
        :param Union[DatabaseCredential,int] databaseCredential: Database Credentials the bot will use
        :param Union[User,int,str] owner: Owner of the bot. Requires Admin to create for other users.
        :return: Bot
        '''

        from . import DatabaseCredential, RedditApp, SentryToken
        additionalParams = {}
        if isinstance(redditApp, RedditApp):
            redditApp = redditApp.id
        if redditApp:
            additionalParams['reddit_app_id'] = redditApp
        if isinstance(sentryToken, SentryToken):
            sentryToken = sentryToken.id
        if sentryToken:
            additionalParams['sentry_token_id'] = sentryToken
        if isinstance(databaseCredential, DatabaseCredential):
            databaseCredential = databaseCredential.id
        if databaseCredential:
            additionalParams['database_credential_id'] = databaseCredential
        if owner:
            additionalParams['owner_id'] = owner
        return _credmgr.post('/bots', data={'app_name': name, **additionalParams})

    def edit(self, **kwargs):
        for key, value in kwargs.items():
            snakeKey = camelToSnake(key)
            if snakeKey in ['reddit_app', 'sentry_token', 'database_credential']:
                newKey = f'{snakeKey}_id'
                kwargs[newKey] = kwargs.pop(key)
        super(Bot, self).edit(**kwargs)


from typing import TypeVar


User = TypeVar("User")
class OwnerMixin:
    _editableAttrs = []

    @property
    def owner(self) -> User:
        user = self._credmgr.user(id=self.ownerId)
        return user


from . import Bot, DatabaseCredential, RedditApp, RefreshToken, SentryToken, User, UserVerification
from .utils import resolveUser
from ..exceptions import InitializationError
from ..mixins import BaseModel


class Paginator:

    @resolveUser()
    def __init__(self, credmgr, model, batchSize=20, limit=None, owner=None):
        '''Initialize a ListGenerator instance.

        :param credmgr: An instance of :class:`.CredentialManager`.
        :param model: A CredentialManager model to list.
        :param int batchSize: The number of items to fetch at a time. If ``batchSize`` is None, it will fetch them 100 at a time. (default: 20).
        :param int limit: The maximum number of items to get.
        :param Union[int, User, str] owner: Owner to filter the items by.
        '''
        self._credmgr = credmgr
        self._model = model(self._credmgr)
        self.batchSize = batchSize
        self.limit = limit
        self.owner = owner
        self._itemsReturned = 0
        self._completed = False
        self._offset = 0
        self._currentItemIndex = None
        self._response = None

    def __iter__(self):
        return self

    def __next__(self):
        if self.limit is not None and self._itemsReturned >= self.limit:
            raise StopIteration()

        if self._response is None or self._currentItemIndex >= len(self._response):
            self._next()

        self._currentItemIndex += 1
        self._itemsReturned += 1
        return self._response[self._currentItemIndex - 1]

    def _next(self):
        if self._completed:
            raise StopIteration()
        params = dict(limit=self.batchSize, offset=self._offset)
        if self.owner:
            params['owner_id'] = self.owner
        self._response = self._credmgr.get(self._model._path, params=params)
        self._currentItemIndex = 0
        if not self._response:
            raise StopIteration()
        if self._response and len(self._response) == self.batchSize:
            self._offset += self.batchSize
        else:
            self._completed = True

class BaseHelper(BaseModel):
    _model = None

    def __call__(self, id=None, name=None, batchSize=20, limit=None, owner=None):
        kwargs = {}
        byName = False
        if isinstance(id, str):
            byName = True
            name = id
            id = None
        if id:
            kwargs['id'] = id
        if name:
            if self._model._canFetchByName:
                kwargs[self._model._nameAttr] = name
            else:
                raise InitializationError(f'Cannot get {self._model.__name__!r} by name')
        if not (id or name):
            return self._model(self._credmgr).listItems(batchSize=batchSize, limit=limit, owner=owner)
        item = self._model(self._credmgr, **kwargs)
        item.fetch(byName)
        return item

class UserHelper(BaseHelper):
    _model = User

    def __call__(self, id=None, name=None, batchSize=20, limit=None):
        kwargs = {}
        byName = False
        if isinstance(id, str):
            byName = True
            name = id
            id = None
        if id:
            kwargs['id'] = id
        if name:
            kwargs['username'] = name
        if not (id or name):
            return User(self._credmgr).listItems(batchSize=batchSize, limit=limit)
        item = User(self._credmgr, **kwargs)
        item.fetch(byName)
        return item

    def create(self, username, password, defaultSettings=None, isAdmin=False, isActive=True, isRegularUser=True, isInternal=False, redditUsername=None) -> User:
        '''Create a new user

        **PERMISSIONS: Admin role is required.**

        :param str username: Username for new user (Example: ```spaz```) (required)
        :param str password: Password for new user (Example: ```supersecurepassword```) (required)
        :param str defaultSettings: Default values to use for new apps (Example: ```{"databaseFlavor": "postgres", "databaseHost": "localhost"}```)
        :param bool isAdmin: Is the user an admin? Allows the user to see all objects and create users (Default: ``False``)
        :param bool isActive: Is the user active? Allows the user to sign in (Default: ``True``)
        :param bool isRegularUser: (Internal use only)
        :param bool isInternal: (Internal use only)
        :param str redditUsername:
        :return: User
        '''

        return self._model._create(self._credmgr, username, password, defaultSettings, isAdmin, isActive, isRegularUser, isInternal, redditUsername)

class BotHelper(BaseHelper):
    _model = Bot

    def create(self, appName, redditApp=None, sentryToken=None, databaseCredential=None, owner=None) -> Bot:
        '''Create a new Bot

        **PERMISSIONS: At least Active user is required.**

        Bots are used for grouping apps into a single request

        :param str appName: Name of the Bot (required)
        :param Union[RedditApp,int] redditApp: Reddit App the bot will use
        :param Union[SentryToken,int] sentryToken: Sentry Token the bot will use
        :param Union[DatabaseCredential,int] databaseCredential: Database Credentials the bot will use
        :param Union[User,int,str] owner: Owner of the bot. Requires Admin to create for other users.
        :return: Bot
        '''

        return self._model._create(self._credmgr, name=appName, redditApp=redditApp, sentryToken=sentryToken, databaseCredential=databaseCredential, owner=owner)

class RedditAppHelper(BaseHelper):
    _model = RedditApp

    def create(self, appName, clientId, userAgent=None, appType='web', redirectUri='https://credmgr.jesassn.org/oauth2/reddit_callback', clientSecret=None, shortName=None,
            appDescription=None, enabled=True, owner=None) -> RedditApp:
        '''Create a new RedditApp

        **PERMISSIONS: At least Active user is required.**

        Reddit Apps are used for interacting with reddit

        :param str appName: (required)
        :param str clientId: Client ID of the Reddit App (required)
        :param str userAgent: User agent used for requests to Reddit's API (required, defaults to user set default, then to 'python:{appName} by /u/{redditUsername}' if currentUser.redditUsername is set or 'python:{appName}' if it is not set)
        :param str appType: Type of the app. One of `web`, `installed`, or `script` (default: 'web')
        :param str redirectUri: Redirect URI for Oauth2 flow. Defaults to user set redirect uri (default: 'https://credmgr.jesassn.org/oauth2/reddit_callback')
        :param str clientSecret: Client secret of the Reddit App
        :param str shortName: Short name of the Reddit App
        :param str appDescription: Description of the Reddit App
        :param bool enabled: Allows the app to be used
        :param int owner: Owner of the app. Requires Admin to create for other users.
        :return: RedditApp
        '''
        if not userAgent:
            redditUsername = self._credmgr.currentUser.redditUsername
            redditUsernameStr = ''
            if redditUsername:
                redditUsernameStr = f' by /u/{redditUsername}'
            userAgent = self._credmgr.getUserDefault('user_agent', f'python:{appName}{redditUsernameStr}')
        return self._model._create(self._credmgr, appName, clientId, userAgent, appType, redirectUri, clientSecret, shortName, appDescription, enabled, owner)

class UserVerificationHelper(BaseHelper):
    _model = UserVerification

    def create(self, userId, redditApp, redditor=None, extraData=None, owner=None) -> UserVerification:
        '''Create a new User Verification

        **PERMISSIONS: At least Active user is required.**

        User Verifications for verifying a redditor with a User ID

        :param str userId: User ID to associate Redditor with (required)
        :param Union[RedditApp,int,str] redditApp: Reddit app the User Verification is for (required)
        :param str redditor: Redditor the User Verification is for
        :param dict extraData: Extra JSON data to include with verification
        :param int owner: Owner of the verification. Requires Admin to create for other users.
        :return: UserVerification
        '''
        return self._model._create(self._credmgr, userId=userId, redditApp=redditApp, redditor=redditor, extraData=extraData, owner=owner)

    def __call__(self, userId=None, redditAppId=None, batchSize=20, limit=None, owner=None):
        kwargs = {}
        if userId:
            kwargs['userId'] = userId
        if redditAppId:
            kwargs['redditAppId'] = redditAppId
        if not userId:
            return self._model(self._credmgr).listItems(batchSize=batchSize, limit=limit, owner=owner)
        item = self._model(self._credmgr, **kwargs)
        item.fetch(True)
        return item

class SentryTokenHelper(BaseHelper):
    _model = SentryToken

    def create(self, appName, dsn, owner=None) -> SentryToken:
        '''Create a new Sentry Token

        **PERMISSIONS: At least Active user is required.**

        Sentry Tokens are used for logging and error reporting in applications

        :param str appName: Name of the Sentry Token (required)
        :param str dsn: DSN of the Sentry Token (required)
        :param Union[User,int,str] owner: Owner of the verification. Requires Admin to create for other users.
        :return: SentryToken
        '''
        return self._model._create(self._credmgr, appName=appName, dsn=dsn, owner=owner)

class DatabaseCredentialHelper(BaseHelper):
    _model = DatabaseCredential

    def create(self, appName, databaseFlavor='postgres', database='postgres', databaseHost='localhost', databasePort=5432, databaseUsername='postgres', databasePassword=None,
            useSSH=False, sshHost=None, sshPort=None, sshUsername=None, sshPassword=None, useSSHKey=False, privateKey=None, privateKeyPassphrase=None, enabled=True,
            owner=None) -> DatabaseCredential:
        '''Create a new Database Credential

        **PERMISSIONS: At least Active user is required.**

        Database Credentials are used for..ya know..databases

        :param str appName: Name of the Database Credential (required)
        :param str databaseFlavor: Type of database, (default: ``postgres``)
        :param str database: Working database to use, (default: ``postgres``)
        :param str databaseHost: Database server address, (default: ``localhost``)
        :param int databasePort: Port the database server listens on, (default: ``5432``)
        :param str databaseUsername: Username to use to connect to the database
        :param str databasePassword: Password to use to connect to the database
        :param bool useSSH: Determines if the database will be connected to through a tunnel
        :param str sshHost: The address of the server that the SSH tunnel will connect to
        :param str sshPort: The port the SSH tunnel will use
        :param str sshUsername: Username for the SSH tunnel
        :param str sshPassword: Password for the SSH tunnel
        :param bool useSSHKey: Allows the credentials to be used
        :param str privateKey: SSH private key. Note: No validation will be performed.
        :param str privateKeyPassphrase: Passphrase for the SSH key
        :param bool enabled: Allows the credentials to be used
        :param Union[User,int,str] owner: Owner of the app. Requires Admin to create for other users.
        :return: DatabaseCredential
        '''
        return self._model._create(self._credmgr, appName=appName, databaseFlavor=databaseFlavor, database=database, databaseHost=databaseHost, databasePort=databasePort,
            databaseUsername=databaseUsername, databasePassword=databasePassword, useSsh=useSSH, sshHost=sshHost, sshPort=sshPort, sshUsername=sshUsername, sshPassword=sshPassword,
            useSshKey=useSSHKey, privateKey=privateKey, privateKeyPassphrase=privateKeyPassphrase, enabled=enabled, owner=owner)

class RefreshTokenHelper(BaseHelper):
    _model = RefreshToken

    def __call__(self, redditor=None, redditAppId=None, batchSize=20, limit=None, owner=None):
        kwargs = {}
        if redditor:
            kwargs['redditor'] = redditor
        if redditAppId:
            kwargs['redditAppId'] = redditAppId
        if not (redditor and redditAppId):
            return self._model(self._credmgr).listItems(batchSize=batchSize, limit=limit, owner=owner)
        item = self._model(self._credmgr, **kwargs)
        item.fetch(True)
        return item


from .base import BaseModel
from .deletable import DeletableMixin
from .editable import EditableMixin
from .owner import OwnerMixin
from .togglable import ToggableMixin


_ = None
del _
from .baseApp import BaseApp